/******************************************************************************
 *
 * info.c - Dump NTFS file informations. Part of the DirTools project.
 *
 * $Revision: 1.3 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#ifdef _WIN32
#include <windows.h>
#else /* _WIN32 */
#include <time.h>
#include <sys/utsname.h>
#endif /* _WIN32 */

#include "info.h"
#include "dt_comment.h"
#include "dt_locale.h"
#include "dt_error.h"
#include "macros.h"
#include "diskio.h"
#include "dt_memory.h"
#ifdef _WIN32
#include "winsid.h"
#endif /* _WIN32 */
#include "md5.h"
#include "nttime.h"
#include "sd_dump.h"

TYPE_COMMENT_BITS CommentFileAttribute[] = {
     {FILE_ATTRIBUTE_READONLY, {"Ro", "Read-Only"}
      }
     ,                          /*  Lecture seule */
     {FILE_ATTRIBUTE_HIDDEN, {"Hi", "Hidden"}
      }
     ,                          /*  Fichier cach */
     {FILE_ATTRIBUTE_SYSTEM, {"Sy", "System"}
      }
     ,                          /*  fichier systme */
     {FILE_ATTRIBUTE_DIRECTORY, {"Di", "Directory"}
      }
     ,                          /*  Rpertoire */
     {FILE_ATTRIBUTE_ARCHIVE, {"Ar", "Archive"}
      }
     ,                          /*  Archive */
     {FILE_ATTRIBUTE_ENCRYPTED, {"En", "Encrypted"}
      }
     ,                          /*  Chiffr */
     {FILE_ATTRIBUTE_NORMAL, {"No", "Normal"}
      }
     ,
     {FILE_ATTRIBUTE_TEMPORARY, {"Te", "Temporary"}
      }
     ,                          /*  Temporaire */
     {FILE_ATTRIBUTE_SPARSE_FILE, {"Sp", "Sparse file"}
      }
     ,
     {FILE_ATTRIBUTE_REPARSE_POINT, {"Re", "Reparse point"}
      }
     ,
     {FILE_ATTRIBUTE_COMPRESSED, {"Co", "Compressed"}
      }
     ,                          /*  Compress */
     {FILE_ATTRIBUTE_OFFLINE, {"Of", "Offline"}
      }
     ,                          /*  Non connect (bande,...) */
     {FILE_ATTRIBUTE_NOT_CONTENT_INDEXED, {"Ni", "Not content indexed"}
      }
     ,
     {0, {NULL, NULL}
      }
};

FILE *FInfoDebug = NULL;
int BDoNotCheckSequence = FALSE;

/**********************************************************************
 * DumpFileName
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 *
 *
 *********************************************************************/
int
DumpFileName (FILE * fOutput, PNTFS_MFT_ENTRY pEntry)
{
     int iErrorLevel = OK_ERROR;
     uint16_t wCount;
     unsigned char *pBuffer = NULL;
     int bInit = FALSE;
     uint64_t ulLength, ulFlags;
     unsigned char byNameSpace = 0;
     uint16_t wNameLength;
     wchar_t wcName[ATTRIBUTE_MAX_NAME_LENGTH + 1];       /* add 1 wchar to print
                                                         * the directory
                                                         * separator */

     wcName[0] = L'\0';
     for (wCount = 0; wCount < pEntry->wAttrCount; wCount++) {
          if (pEntry->pAttr[wCount].dwType == pEntry->pVolume->dwAtFileName) {
               pBuffer =
                    GetStreamBuffer (pEntry, &pEntry->pAttr[wCount],
                                     &ulLength);

               if (ulLength >= 0x42) {
                    byNameSpace = DT_GETU8 (pBuffer + 0x41);
                    if ((!bInit) || (byNameSpace & 1))
                         /* We prefer Win32 name spaces i.e. byNameSpace== 1
                          * or 3 */
                    {
                         bInit = TRUE;
                         ulFlags = DT_GETU64 (pBuffer + 0x38);
                         wNameLength = (uint16_t) DT_GETU8 (pBuffer + 0x40);
                         if (((wNameLength * 2) + 0x42) <= ulLength)
                         {
                              dt_get_unicode (pBuffer + 0x42, wcName,
                                           (uint32_t) wNameLength);
                              if (ulFlags & 0x10000000) {
                                   wcName[wNameLength] = L'\\';
                                   wNameLength++;
                              }
                              wcName[wNameLength] = L'\0';
                         }
                    }           /* if ((!bInit) || (byNameSpace & 1)) */
               } /* if (ulLength >= 0x42) */
               else {
                    ALARM (TA_PROG_ERR, "truncated stream - Size : %" PRIu64,
                           ulLength);
                    iErrorLevel = SIMPLE_ERROR;
               }

               if (ReleaseStreamBuffer (&pEntry->pAttr[wCount], pBuffer) ==
                   FATAL_ERROR)
                    return FATAL_ERROR;

               if (byNameSpace & 1)
                    /* We have a Win32 name, get out of the loop */
                    break;

          }                     /* if (pEntry->pAttr[wCount].dwType == ... */
     }                          /* for (wCount = 0; wCount < pEntry->... */
     if (bInit)
          /* Some information has been found */
	  dt_fputwcs (fOutput, wcName);

     return iErrorLevel;
}

/**********************************************************************
 * WalkBack
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
int
WalkBack (FILE * fOutput, PNTFS_MFT_ENTRY pEntry, int bAnyFile, int bCompact)
{
     int iErrorLevel = OK_ERROR;
     PNTFS_ATTRIBUTE pFileAttr = NULL;
     unsigned char *pFileBuffer = NULL;
     uint64_t ulFileLength, ulParent;
     PNTFS_MFT_ENTRY pParent = NULL;
     uint16_t wParent;

     if (FInfoDebug)
          fprintf (FInfoDebug,
                   "%s: dealing with base record #%" PRIu64 "\n",
                   __func__, pEntry->ulNumber);

     pFileAttr = FindAttribute (pEntry, pEntry->pVolume->dwAtFileName, NULL);
     if (!pFileAttr) {
          ALARM (TA_PROG_ERR,
                 "unable to find any filename\nBase record #%" PRIu64,
                 pEntry->ulNumber);
          iErrorLevel = SIMPLE_ERROR;
     } else {                   /* if (! pFileAttr) */

          pFileBuffer = GetStreamBuffer (pEntry, pFileAttr, &ulFileLength);
          if (ulFileLength >= 8) {
               ulParent = DT_GETU48 (pFileBuffer);
               wParent = DT_GETU16 (pFileBuffer + 6);
               if (FInfoDebug)
                    fprintf (FInfoDebug,
                             "%s: found parent directory #%" PRIu64
                             " (seq %" PRIu16 ")\n", __func__, ulParent,
                             wParent);
               if (bAnyFile || pEntry->bDirectory) {
                    if (ulParent == 5) {
                         /* . (root directory) found */
#ifdef _WIN32
                         fprintf (fOutput, "%c:\\", pEntry->pVolume->tDrive);
#else /* _WIN32 */
                         fprintf (fOutput, "%s:\\", pEntry->pVolume->file);
#endif /* _WIN32 */
                    } else {    /* if (ulParent == 5) */

                         switch (InitEntry
                                 (pEntry->pVolume, &pParent, ulParent,
                                  FALSE)) {
                         case SIMPLE_ERROR:
                              iErrorLevel = SIMPLE_ERROR;
                              break;
                         case FATAL_ERROR:
                              return FATAL_ERROR;
                              break;
                         }
                         if (BDoNotCheckSequence
                             || pParent->wSequence == wParent) {
                              switch (WalkBack
                                      (fOutput, pParent, FALSE, bCompact)) {
                              case SIMPLE_ERROR:
                                   iErrorLevel = SIMPLE_ERROR;
                                   break;
                              case FATAL_ERROR:
                                   return FATAL_ERROR;
                                   break;
                              }
                         } else if (FInfoDebug) {
                              fprintf (FInfoDebug,
                                       "%s: parent #%" PRIu64
                                       ", unexpected sequence %" PRIu16
                                       " (expected %" PRIu16 ")\n", __func__,
                                       ulParent, pParent->wSequence, wParent);
                         }
                         if (FreeEntry (pParent) == FATAL_ERROR)
                              return FATAL_ERROR;
                    }           /* if (ulParent == 5) */
                    switch (DumpFileName (fOutput, pEntry)) {
                    case SIMPLE_ERROR:
                         iErrorLevel = SIMPLE_ERROR;
                         break;
                    case FATAL_ERROR:
                         return FATAL_ERROR;
                         break;
                    }
               } /* if (bAnyFile || pEntry->bDirectory) */
               else if (FInfoDebug)
                    fprintf (FInfoDebug,
                             "%s: expected parent #%" PRIu64
                             " is not a directory\n", __func__, ulParent);
          } else {              /* if (ulFileLength >= 0x40) */

               ALARM (TA_PROG_ERR,
                      "WalkBack : corrupted filename stream - "
                      "Base record #%" PRIu64, pEntry->ulNumber);
               iErrorLevel = SIMPLE_ERROR;
          }                     /* if (ulFileLength >= 0x40) */
          if (ReleaseStreamBuffer (pFileAttr, pFileBuffer) == FATAL_ERROR)
               return FATAL_ERROR;
     }                          /* if (! pFileAttr) */

     return iErrorLevel;
}

/**********************************************************************
 * DumpFullPath
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
int
DumpFullPath (FILE * fOutput, PNTFS_MFT_ENTRY pEntry, int bCompact)
{
     int iErrorLevel = OK_ERROR;

     if (!bCompact)
          fprintf (fOutput, "Full path name : ");

     switch (WalkBack (fOutput, pEntry, TRUE, bCompact)) {
     case SIMPLE_ERROR:
          iErrorLevel = SIMPLE_ERROR;
          break;
     case FATAL_ERROR:
          return FATAL_ERROR;
          break;
     }

     fprintf (fOutput, (bCompact ? "|" : "\n"));

     return iErrorLevel;
}

/**********************************************************************
 * DumpFileTime
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
void
DumpFileTime (FILE * fOutput, uint64_t ulTime, int bCompact)
{

     time_prn (fOutput, ulTime);
     fprintf (fOutput, (bCompact ? "|" : "\n"));
}

/**********************************************************************
 * DumpStandard
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 *
 *********************************************************************/
int
DumpStandard (FILE * fOutput, PNTFS_MFT_ENTRY pEntry,
              PNTFS_ATTRIBUTE pAttr, int bCompact)
{
     int iErrorLevel = OK_ERROR;
     uint64_t ulLength, ulCTime, ulMTime, ulRMTime, ulATime, ulQuota, ulUsn;
     unsigned char *pBuffer = NULL;
     uint32_t dwFlags, dwMVersion, dwCVersion, dwClass, dwOwner;

     pBuffer = GetStreamBuffer (pEntry, pAttr, &ulLength);
     if (ulLength >= 0x30) {
          ulCTime = DT_GETU64 (pBuffer);
          ulMTime = DT_GETU64 (pBuffer + 0x8);
          ulRMTime = DT_GETU64 (pBuffer + 0x10);
          ulATime = DT_GETU64 (pBuffer + 0x18);
          dwFlags = DT_GETU32 (pBuffer + 0x20);
          /* Sometimes the flag looks incomplete... */
          if (pEntry->bDirectory)
               dwFlags |= FILE_ATTRIBUTE_DIRECTORY;

          if (!bCompact)
               fprintf (fOutput, "File attributes : ");
          DumpBits (fOutput, (uint32_t) dwFlags, CommentFileAttribute,
                    bCompact);

          if (!bCompact)
               fprintf (fOutput, "Creation time : ");
          DumpFileTime (fOutput, ulCTime, bCompact);

          if (!bCompact)
               fprintf (fOutput, "Attribute modification time : ");
          DumpFileTime (fOutput, ulMTime, bCompact);

          if (!bCompact)
               fprintf (fOutput, "Last write time : ");
          DumpFileTime (fOutput, ulRMTime, bCompact);

          if (!bCompact)
               fprintf (fOutput, "Last access time : ");
          DumpFileTime (fOutput, ulATime, bCompact);

          if (!bCompact) {
               dwMVersion = DT_GETU32 (pBuffer + 0x24);
               dwCVersion = DT_GETU32 (pBuffer + 0x28);
               dwClass = DT_GETU32 (pBuffer + 0x2C);
               if (dwMVersion)
                    fprintf (fOutput, "Version : %" PRIu32 "/%" PRIu32 "\n",
                             dwCVersion, dwMVersion);
               fprintf (fOutput, "Class Id : %" PRIu32 "\n", dwClass);
          }
          if (pEntry->pVolume->wMajorVersion >= 3) {
               if (ulLength >= 0x48) {
                    pEntry->dwSecId = DT_GETU32 (pBuffer + 0x34);
                    pEntry->bSecIdFound = TRUE;
                    if (!bCompact) {
                         dwOwner = DT_GETU32 (pBuffer + 0x30);
                         ulQuota = DT_GETU64 (pBuffer + 0x38);
                         ulUsn = DT_GETU64 (pBuffer + 0x40);

                         fprintf (fOutput, "Security Id : %" PRIu32 "\n",
                                  pEntry->dwSecId);
                         fprintf (fOutput, "Quota Owner Id : %" PRIu32 "\n",
                                  dwOwner);
                         if (ulQuota)
                              fprintf (fOutput,
                                       "Quota charged : %" PRIu64 "\n",
                                       ulQuota);
                         fprintf (fOutput,
                                  "Update Sequence Number : %" PRIu64 "\n",
                                  ulUsn);
                    }
               } else {
                    ALARM (TA_PROG_WARN,
                           "pre-2k standard stream - Size : %" PRIu64
                           " - Base record %" PRIu64, ulLength,
                           pEntry->ulNumber);
               }
          }
     } else {
          ALARM (TA_PROG_ERR,
                 "truncated stream - Size : %" PRIu64
                 " - Base record %" PRIu64, ulLength, pEntry->ulNumber);
          iErrorLevel = SIMPLE_ERROR;
          fprintf (fOutput, "||||");
     }

     if (ReleaseStreamBuffer (pAttr, pBuffer) == FATAL_ERROR)
          return FATAL_ERROR;

     return iErrorLevel;
}

/**********************************************************************
 * Dump_SD_Security
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
int
Dump_SD_Security (FILE * fOutput, PNTFS_MFT_ENTRY pEntry,
                  PNTFS_ATTRIBUTE pAttr, int bCompact)
{
     int iErrorLevel = OK_ERROR;
     uint64_t ulLength;
     unsigned char *pBuffer = NULL;

     pBuffer = GetStreamBuffer (pEntry, pAttr, &ulLength);

     iErrorLevel = DumpBufferSecurity (fOutput, pEntry->ulNumber, pBuffer,
                                       ulLength, bCompact);

     if (ReleaseStreamBuffer (pAttr, pBuffer) == FATAL_ERROR)
          return FATAL_ERROR;

     return iErrorLevel;
}

/**********************************************************************
 * Dump_SDS_Security
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
int
Dump_SDS_Security (FILE * fOutput, PNTFS_MFT_ENTRY pEntry, int bCompact)
{
     uint64_t ulLength;
     uint32_t dwIndex = 0;
     static unsigned char *pBuffer = NULL;

     if (pBuffer == NULL) {
          if (pEntry->pVolume->secList.dwMaxSize == 0)
               return SIMPLE_ERROR;
          pBuffer =
               DT_CALLOC (unsigned char, pEntry->pVolume->secList.dwMaxSize,
                          "pEntry->pVolume->secList.dwMaxSize - Base record #%"
                          PRIu64, pEntry->ulNumber);
     }

     while (dwIndex < pEntry->pVolume->secList.dwListSize) {
          if (pEntry->pVolume->secList.pSecInfo[dwIndex].dwSecId ==
              pEntry->dwSecId) {
               ulLength = pEntry->pVolume->secList.pSecInfo[dwIndex].dwSize;
               if (ReadStreamBytes
                   (pEntry->pVolume, pEntry->pVolume->pSecureData,
                    pEntry->pVolume->secList.pSecInfo[dwIndex].ulSDSOffset,
                    &ulLength, pBuffer) == FALSE) {
                    ALARM (TA_PROG_ERR,
                           "security id %" PRIu32 " not read (%" PRIu64
                           ") - Base record : %" PRIu64, pEntry->dwSecId,
                           ulLength, pEntry->ulNumber);
                    return SIMPLE_ERROR;
               }

               return DumpBufferSecurity (fOutput, pEntry->ulNumber, pBuffer,
                                          ulLength, bCompact);
          }
          dwIndex++;
     }
     ALARM (TA_PROG_ERR, "security id %" PRIu32 " not found - Base record : %"
            PRIu64, pEntry->dwSecId, pEntry->ulNumber);

     return SIMPLE_ERROR;
}

/**********************************************************************
 * DumpMd5
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
int
DumpMd5 (FILE * fOutput, PNTFS_MFT_ENTRY pEntry, int bCompact)
{
     int bCorrect = TRUE;
     MD5_CTX mdContext;
     uint16_t wCount;
     unsigned char ucDigest[16];

     MD5Init (&mdContext);

     for (wCount = 0; wCount < pEntry->wAttrCount; wCount++) {
          if (pEntry->pAttr[wCount].dwType == pEntry->pVolume->dwAtData)
               bCorrect = ProcessStream (pEntry, &pEntry->pAttr[wCount],
                                         &ProcessMd5, &mdContext);

          if (!bCorrect) {
               ALARM (TA_PROG_ERR,
                      "error while processing MD5 - Base record #%" PRIu64,
                      pEntry->ulNumber);
               break;
          }
     }

     MD5Final (ucDigest, &mdContext);

     if (!bCompact)
          fprintf (fOutput, "MD5 hash value : ");

     for (wCount = 0; wCount < 16; wCount++)
          fprintf (fOutput, "%02x", ucDigest[wCount]);

     fprintf (fOutput, (bCompact ? "|" : "\n"));

     return bCorrect;
}

/**********************************************************************
 * DumpFileInfo
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
int
DumpFileInfo (FILE * fOutput, PNTFS_MFT_ENTRY pEntry, int bMd5, int bCompact)
{
     int iErrorLevel = OK_ERROR, iErrorSDS = OK_ERROR;
     PNTFS_ATTRIBUTE pAttr = NULL;
     uint16_t wDataAttr, wCount;

     /* Print index of the file in the MFT */
     fprintf (fOutput,
              (bCompact ? "%" PRIu64 "|" : "MFT record number : %" PRIu64
               "\n"), pEntry->ulNumber);

     /* Print name of the file */
     switch (DumpFullPath (fOutput, pEntry, bCompact)) {
     case SIMPLE_ERROR:
          iErrorLevel = SIMPLE_ERROR;
          break;
     case FATAL_ERROR:
          return FATAL_ERROR;
          break;
     }

     /* Compute MD5 hash */
     if (bMd5) {
          if (!DumpMd5 (fOutput, pEntry, bCompact))
               iErrorLevel = SIMPLE_ERROR;
     } else if (bCompact)
          fprintf (fOutput, "|");

     /* Print number of hard links */
     if (pEntry->wRefCount)
          fprintf (fOutput,
                   (bCompact ? "%u|" : "Number of hard links : %u\n"),
                   pEntry->wRefCount);
     else
          fprintf (fOutput, (bCompact ? "0|" : "Erased file\n"));

     /* Print "primary" data stream size (if any) = file size
        pAttr = FindAttribute(pEntry, pEntry->pVolume->dwAtData, L""); */
     pAttr = FindAttribute (pEntry, pEntry->pVolume->dwAtData, L'\0');
     if (pAttr)
          fprintf (fOutput,
                   (bCompact ? "%" PRIu64 "|" : "Data stream size : %" PRIu64
                    "\n"), pAttr->ulStreamSize);
     else if (bCompact)
          fprintf (fOutput, "|");

     /* Print standard attributes, i.e. timestamps and flags */
     pAttr = FindAttribute (pEntry, pEntry->pVolume->dwAtStdInfo, NULL);
     if (pAttr) {
          switch (DumpStandard (fOutput, pEntry, pAttr, bCompact)) {
          case SIMPLE_ERROR:
               iErrorLevel = SIMPLE_ERROR;
               break;
          case FATAL_ERROR:
               return FATAL_ERROR;
               break;
          }
     }
/* FIXME should be in DumpStandard */
     else if (bCompact)
          fprintf (fOutput, "||||");

     /* Print the number of data streams */
     wDataAttr = 0;
     for (wCount = 0; wCount < pEntry->wAttrCount; wCount++) {
          if (pEntry->pAttr[wCount].dwType == pEntry->pVolume->dwAtData)
               wDataAttr++;
     }
     fprintf (fOutput,
              (bCompact ? "%u|" : "Number of data stream : %u\n"), wDataAttr);

     /* Print the security descriptor */
     iErrorSDS = SIMPLE_ERROR;
     if (pEntry->pVolume->wMajorVersion >= 3) {
          if (pEntry->bSecIdFound) {
               iErrorSDS = Dump_SDS_Security (fOutput, pEntry, bCompact);
               if (iErrorSDS == SIMPLE_ERROR)
                    ALARM (TA_PROG_ERR, "security id %" PRIu32
                           " not found, trying SD stream - Base record : %"
                           PRIu64, pEntry->dwSecId, pEntry->ulNumber);
          } else
               iErrorSDS = SIMPLE_ERROR;
     }
     if (iErrorSDS == SIMPLE_ERROR) {
          pAttr = FindAttribute (pEntry, pEntry->pVolume->dwAtSecurity, NULL);
          if (pAttr) {
               switch (Dump_SD_Security (fOutput, pEntry, pAttr, bCompact)) {
               case SIMPLE_ERROR:
                    iErrorLevel = SIMPLE_ERROR;
                    break;
               case FATAL_ERROR:
                    return FATAL_ERROR;
                    break;
               }
          }
/* FIXME should be in Dump_SD_Security */
          else if (bCompact)
               fprintf (fOutput, "||||");
     }

     /* Print (if any) alternate data streams name and size */
     wDataAttr = 0;
     for (wCount = 0; wCount < pEntry->wAttrCount; wCount++) {
          if ((pEntry->pAttr[wCount].dwType == pEntry->pVolume->dwAtData) &&
              (pEntry->pAttr[wCount].wcName[0] != L'\0')) {
               if (!bCompact)
                    fprintf (fOutput, "Alternate data stream #%u", wDataAttr);
	       if (! bCompact)
	            fputs(" Name: ", fOutput);
	       dt_fputwcs(fOutput, pEntry->pAttr[wCount].wcName);
               fprintf (fOutput,
                         (bCompact ? "|%" PRIu64 "|" :
                          "\n Size : %" PRIu64 "\n"),
                         pEntry->pAttr[wCount].ulStreamSize);
               wDataAttr++;
          }
     }

     fprintf (fOutput, "\n");

     return iErrorLevel;
}
