/******************************************************************************
 *
 * utility.c - program options support. Part of the DirTools project.
 *
 * $Revision: 1.1.1.1 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <stdint.h>
#ifdef _WIN32
#include <io.h>
#else /* _WIN32 */
#include <string.h>
#endif /* _WIN32 */

#include "utility.h"


/**********************************************************************
 * CreateOutputFile : create a new file for output, check if already
 *                    exists
 *
 * - [in] pFileName : pointer to requested file name
 *
 * returns pointer to file descriptor in case of success, NULL otherwise
 *
 *********************************************************************/
FILE *
CreateOutputFile (const char *pFileName)
{
     int iFile;
     FILE *fResult;

     if ((iFile = open (pFileName, O_CREAT | O_EXCL | O_WRONLY)) < 0) {
          if (errno == EEXIST)
               fprintf (stderr, "%s already exists\n", pFileName);
          else
               fprintf (stderr, "Can't create %s\n", pFileName);
          return NULL;
     }

     fResult = fdopen (iFile, "w");

     return fResult;
}

/**********************************************************************
 * integer_arg : transform a string into an integer (taken from strings)
 *
 * - [in] pString : pointer to string to transform
 *
 * returns an unsigned 64bits integer corresponding to the string
 *        (no overflow check)
 *********************************************************************/
uint64_t
integer_arg (const char *pString)
{
     uint64_t ulValue = 0;
     int wBase = 10;
     const char *pChar = pString;
     int wCurrent;

     if (*pChar != '0')
          wBase = 10;
     else if (*++pChar == 'x') {
          wBase = 16;
          pChar++;
     } else
          wBase = 8;

     ulValue = 0;

     while (((wCurrent = *pChar++) >= '0' && wCurrent <= '9') ||
            ((wBase == 16) && (wCurrent & ~32) >= 'A'
             && (wCurrent & ~32) <= 'F')) {
          ulValue *= wBase;
          if (wCurrent >= '0' && wCurrent <= '9')
               ulValue += wCurrent - '0';
          else
               ulValue += (wCurrent & ~32) - 'A' + 10;
     }

     return ulValue;
}
